from decimal import Decimal
import numpy as np
from scipy.optimize import leastsq
from  scipy import signal
import re
import math
import os
import numbers
import collections
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

class ArgError(Exception):
  pass

""" Module containing many small methods to do various usefull things

check_kwargs -- takes a list of expected keywords and actual keywords.
  Returns an error if an actual keyword is not expected. Does nothing if an
  expected keyword is not present. Could be improved with a list of mandatory
  and a list of optional keuword.
check_and_mkdir -- Check/create all the directoryies in a file address.
clean_white_spaces -- Replace all tabs or repeated spaces by a single space
in given list of strings.
stround -- Rounds a number-like string to given number of decimals
 """

def check_kwargs(expected, kwargs):
  for key in kwargs:
    if not key in expected:
      raise ArgError("Unknown keyword argument",key)

def check_and_mkdir(path, verbose = 0, skip = ""):
  """ Check if all the directoryies in a file address exist.
  
  Creates them if possible.
  If last item of the path is a file, it must contain a '.', otherwise
  will be created as a directory.
  """
  
  if "." in path.split("/")[-1]:
    path = path.split("/")[:-1]
    path = str("/").join(path)
  if not os.path.exists(path):
    try:
      os.makedirs(path)
      if verbose > 0:
        print(skip, "Info in toolbox: created directory", path)
    except Exception as e:
      print("Error in toolbox::check_and_mkdir: %s" % str(e))
      return False
  return True

def clean_white_spaces(lines):
  """ Replace all tabs or repeated spaces by a single space in given list.
  
  Argument:
    lines -- List of string objects. Mutates it, does not return anything.
  """
  
  for iline in range(len(lines)):
    lines[iline] = lines[iline].replace('\t',' ')
    while '  ' in lines[iline]:
      lines[iline] = lines[iline].replace('  ',' ')

def stround(string, decimals):
  """ Rounds a number-like string to given number of decimals.
  
  Arguments:
    string -- A number-like string.
    decimals -- Precision desired.
  Returns a string.
  """
  
  return str(round(Decimal(string),decimals))

def general_fit(method,x,y,par,range_p = []):
  """ Method to call to fit points with the given function.
  
  Arguments:
    method -- String representation of the function. Must be callable by eval,
      or be 'gauss' or 'doublegauss'. Can also be a python function, see
      e_general_fit for more details.
    x -- List of x coordinates.
    y -- List of y coordinates. Must be same size as x.
    par -- Initial values of the function parameters.
    range_p -- Upper and lower limits allowed for each parameters.
      If specified, must be a list of pair that is twice the size of par.
      If not specified, parameters values will not be limited.
      (Default [])
  
  Returns the list of fitted parameters values, the chisquare of the fit,
  and a pair of fitted xs and ys values covering the same range than the given
  xs and ys.
  """
  if len(par) == 0:
    raise ArgError("Can not fit without parameters.")
  if len(range_p) != 0 and len(range_p) != 2*len(par):
    raise ArgError("The size of the list of parameters limits must be twice "\
                  " as long as the size of the list of parameters.")
  if len(par) > len(x):
    raise ArgError("You gave me less x than parameters, "\
                   "can not fit in those conditions!")
  
  p, cov, infodict, mesg, ier = leastsq(e_general_fit, par[:], 
                                        args=(method,x, y, range_p[:]),
                                        maxfev=100000,
                                        full_output=1)
  xx = np.linspace(x[0], x[-1], 100)
  ffit = []
  if not callable(method):
    ffit = function_from_string(p, xx, method)
  else:
    ffit = method(xx, *p)
  ss_err = (infodict['fvec']**2).sum()
  ss_tot = ((y - np.array(y).mean())**2).sum()
  chisquare = 1 - (ss_err / ss_tot)
  
  return p, chisquare, [xx, ffit]

def e_general_fit(p, method, x, y, range_p = []):
  """ Difference between given method and parameters to data.

  Called by general_fit.
  Argument:
    p -- List of parameters to use.
    method -- string of the function to compare data to, or a python function.
              If a python function is used, it must be of the form 
              func(x, par0, par1, par2...)
    x -- A single x coordinates or a list of coordinates.
    y -- A single y coordinates or a list of coordinates.
    range_p -- see general_fit method.
  
  Returns the difference(s) between given y and the y returned by the function
  at the point(s) x.
  """
  if len(range_p) != 2*len(p) and len(range_p) != 0:
    raise ArgError("The size of the list of parameters limits must be twice "\
                  " as long as the size of the list of parameters.")
  if len(range_p) > 0:
    for ip in range(len(p)):
      par = p[ip]
      ran = range_p[2*ip:2*ip+2]
      if not (par > ran[0] and par < ran[1]):
        ret = [1e9]*len(x) #  Returns a high enough value to make the fit bad
        return ret
  if isinstance(method, str):
    return function_from_string(p,x,method) - y
  elif callable(method):
    if not type(x) is np.array:
      return method(np.array(x), *p) - y
    else:
      return method(x, *p) - y
  else:
    raise TypeError("Given method is neither a string nor a function.")

def function_from_string(p,x,method):
  """ Evaluate a function-like string at a given point and parameters.
  
  Arguments:
    p -- parameters.
    x - point to compute the function at. Can be a list.
    method -- string of the function to evaluate. Can also be 'gauss' or
      'doublegauss'.
    
  Returns the value of the function at x with parameters p. If x is an
  itarable, will return a np.array of values.
  """
  
  if method == "gauss":
    return gauss(p,x)
  elif method == "doublegauss":
    return double_gauss(p,x)
  else:
    if isinstance(x,numbers.Number):
      return eval(method)
    elif isinstance(x,collections.Iterable):
      result = []
      Xs = x[:]
      for x in Xs:
        result.append(eval(method))
      return np.array(result)
    else:
      raise ValueError("x has wrong format. Must be a number or an iterable.")

#def replace_par_by_number(par,method):
#  npar = len(par)
#  previous_c_is_p = False
#  previous_c_is_open = False
#  in_param = False
#  for c in method:
#    if c == "p" and not in_param:
#      

def gauss(p, x):
  """ Gaussian function.
  
  Arguments:
    p -- List of parameters sigma, mean and integral.
    x -- Value to compute the gaussian at.  Can be np.array or single value.
  
  Returns the value of the gaussian at point(s) x.
  """
  
  result = (p[0] * (1/np.sqrt(2*math.pi*(p[2]**2)))
                 * np.exp(-(x-p[1])**2 / (2*p[2]**2)))
  return result

def double_gauss(p, x):
  """ Double gaussian function.
  
  Arguments:
    p -- List of parameters sigma0, mean0, integral0, sigma1, mean1, integral1
    x -- Value to compute the gaussian at.  Can be np.array or single value.
  
  Returns the value of the double gaussian at point(s) x.
  """
  
  return (p[0] * (1/np.sqrt(2*math.pi*(p[2]**2)))
               * np.exp(-(x-p[1])**2 / (2*p[2]**2))
        + p[3] * (1/np.sqrt(2*math.pi*(p[5]**2)))
               * np.exp(-(x-p[4])**2 / (2*p[5]**2)))
  
def find_local_max(x,y,arange, bin_width = 0, n_maxs = 0):
  """ Finds local maxima of given data.
  
  Uses scipy.signal.find_peaks_cwt.
  Arguments:
    x -- x coordinates.
    y -- y coordinates.
    bin_width -- distance between two points on x axis. If 0, will use the
      first 2 values of x ton compute it, but then only makes sens is all the
      given xs are evenly spaced. (Default 0).
    n_maxs -- Number of maxima to find. If 1 or less, will only find one.
      (Default 0).
  """

  if len(x) != len(y):
    raise ArgError("x and y shoud have same size.")
  if len(x) < 2:
    raise ValueError("Should have at least two points in given data.")
  bin_width = x[1] - x[0]
  Break = False
  if n_maxs <= 0:
    n_maxs = -1
    Break = True
  peak_width = 4
  ntrials = 0
  maxs = []
  xmaxs = []
  ymaxs = []
  try:
    while len(maxs) != n_maxs and ntrials < 20:
      myrange = np.linspace(peak_width*bin_width,
                            float(arange[-1]-arange[0]),
                            10)
      maxs = signal.find_peaks_cwt(y,myrange)
      xmaxs = [ x[m] for m in maxs ]
      ymaxs = [ y[m] for m in maxs ]
      peak_width = peak_width*1.5
      ntrials += 1
      if Break:
        break
    print("    Local maxs found : ",xmaxs)
  except:
    warnings.warn("Failed",Warning)
  return xmaxs, ymaxs
